@@ -7,9 +7,10 @@ const int MAX_BUF=10240;
 #define NEEDS_VAR   "needs"
 #define SECTION_VAR "section"
 #define SORT_VAR    "sort"
-//#define SUBSECTION_VAR "subsection"
 #define TITLE_VAR   "title"
 #define HOTKEY_VAR  "hotkey"
+#define PRIVATE_ENTRYCOUNT_VAR "__PRIVATE__ENTRYCOUNT"
+#define PRIVATE_ENTRYINDEX_VAR "__PRIVATE__ENTRYINDEX"
 
 //conditionals in the menuentry files (like "package(vi)")
 #define COND_PACKAGE "package"
@@ -17,3 +18,4 @@ const int MAX_BUF=10240;
 
 #define COMPAT_MODE  "menu-1"
 
+
