@@ -16,26 +16,7 @@
 
 int show_time=0, verbose=0, dodebug=0;
 
-prefix_func    prefix_obj;
-ifroot_func    ifroot_obj;
-
-print_func     print_obj;
-ifempty_func   ifempty_obj;
-ifnempty_func  ifnempty_obj;
-iffile_func    iffile_obj;
-ifelse_func    ifelse_obj;
-ifeq_func      ifeq_obj;
-ifneq_func     ifneq_obj;
-ifeqelse_func  ifeqelse_obj;
-
-cond_surr_func cond_surr_obj;
-esc_func       esc_obj;
-escwith_func   escwith_obj;
-escfirst_func  escfirst_obj;
-replacewith_func replacewith_obj;
-cppesc_func    cppesc_obj;
-parent_func    parent_obj;
-basename_func  basename_obj;
+map <String, func *, less<String> > func_data;
 
 menuentry menu;
 configinfo *config;
@@ -61,7 +42,46 @@ void usage(){
       <<"     -showtime : show timeing information"<<endl;
   exit(1);
 }
-bool empty_string(const String &s){
+
+int store_func(func *f){
+  func_data[f->name()]=f;
+}
+int add_functions(){
+  store_func(new prefix_func);
+  store_func(new ifroot_func);
+
+  store_func(new print_func);
+  store_func(new add_func);
+  store_func(new sub_func);
+  store_func(new mult_func);
+  store_func(new div_func);
+  store_func(new ifempty_func);
+  store_func(new ifnempty_func);
+  store_func(new iffile_func);
+  store_func(new ifelsefile_func);
+  store_func(new ifelse_func);
+  store_func(new catfile_func);
+
+  store_func(new ifeq_func);
+  store_func(new ifneq_func);
+  store_func(new ifeqelse_func);
+
+  store_func(new cond_surr_func);
+  store_func(new esc_func);
+  store_func(new escwith_func);
+  store_func(new escfirst_func);
+  store_func(new tolower_func);  
+  store_func(new toupper_func);
+  store_func(new replacewith_func);
+  store_func(new cppesc_func);
+  store_func(new parent_func);
+  store_func(new basename_func);
+  store_func(new entrycount_func);
+  store_func(new entryindex_func);
+}
+
+
+bool empty_String(const String &s){
   if(s.length()&&(s!="none"))
     return false;
   else
@@ -107,7 +127,7 @@ void closegenoutput(){
   delete genoutputfile;
 }
 void genoutput(const String &s,
-	       map<string, string, less<string> > &v){
+	       map<String, String, less<String> > &v){
   
   static String outputname="";
   const char *t;
@@ -139,9 +159,9 @@ cat_str::cat_str(parsestream &i){
   c=i.get_char();
   i.put_back(c);
   try{
-    while((c=i.get_char())){
-      if(isspace(c))
-	continue;
+    while(1){
+      i.skip_space();
+      c=i.get_char();
       i.put_back(c);
       if(isalpha(c))
 	v.push_back(new func_str(i));
@@ -156,13 +176,13 @@ cat_str::cat_str(parsestream &i){
       else if(c=='\0')
 	break;
       else
-	throw char_unexpected(c);
+	throw char_unexpected(&i, c);
     }
-  }catch(endoffile){};
+  }catch(endofline p){};
 }
 
 const_str::const_str(parsestream &i){
-  data=i.get_stringconst();
+  data=i.get_Stringconst();
 }
 
 var_str::var_str(parsestream &i){
@@ -172,45 +192,31 @@ var_str::var_str(parsestream &i){
 
 func_str::func_str(parsestream &i){
   char c;
-  string name;
-  
+  String name;
+  map <String, func *, less<String> >::iterator j;
+
   name=i.get_name();
-  if(name=="prefix")          f=&prefix_obj;
-  else if(name=="ifroot")     f=&ifroot_obj;
-  else if(name=="print")      f=&print_obj;
-  else if(name=="ifempty")    f=&ifempty_obj;
-  else if(name=="ifnempty")   f=&ifnempty_obj;
-  else if(name=="iffile")     f=&iffile_obj;
-  else if(name=="esc")        f=&esc_obj;
-  else if(name=="ifelse")     f=&ifelse_obj;
-  else if(name=="ifeq")       f=&ifeq_obj;
-  else if(name=="ifneq")      f=&ifneq_obj;
-  else if(name=="ifeqelse")   f=&ifeqelse_obj;
-  else if(name=="cond_surr")  f=&cond_surr_obj;
-  else if(name=="escwith")    f=&escwith_obj;
-  else if(name=="escfirst")   f=&escfirst_obj;
-  else if(name=="replacewith")f=&replacewith_obj;
-  else if(name=="cppesc")     f=&cppesc_obj;
-  else if(name=="parent")     f=&parent_obj;
-  else if(name=="basename")   f=&basename_obj;
-  
-  else{
-    cerr<<"Unknown function "+name<<endl;
-    throw informed_fatal();
-  }
+  j=func_data.find(name);
+  if(j==func_data.end()){
+    throw unknown_function(&i);
+  } else
+    f=(*j).second;
+
   i.skip_space();
   i.skip_char('(');
   do{
+    i.skip_space();
     c=i.get_char();
     if(c==')')
       break;
     i.put_back(c);
     args.push_back(new cat_str(i));
+    i.skip_space();
   } while((c=i.get_char())&&(c==','));
   i.put_back(c);
   i.skip_char(')');
   if(args.size()!=(unsigned int)f->nargs())
-    throw narg_mismatch(&(name[0]));
+    throw narg_mismatch(&i, &(name[0]));
 }
 
 
@@ -219,11 +225,11 @@ func_str::func_str(parsestream &i){
 //
 
 ostream &const_str::output(ostream &o, 
-			   map<string, string, less<string> > &/*menuentry*/){
+			   map<String, String, less<String> > &/*menuentry*/){
   return o<<data;
 }
 
-String cat_str::soutput(map<string, string, less<string> > &menuentry){
+String cat_str::soutput(map<String, String, less<String> > &menuentry){
   char buf[MAX_BUF];
   vector <str * >::iterator i;
   ostrstream s(buf,sizeof(buf));
@@ -235,46 +241,46 @@ String cat_str::soutput(map<string, string, less<string> > &menuentry){
   return String(buf);
 }
 ostream &cat_str::output(ostream &o, 
-			 map<string, string, less<string> > &menuentry){
+			 map<String, String, less<String> > &menuentry){
   o<<soutput(menuentry);
   return o;
 }
-void cat_str::output(map<string, string, less<string> > &menuentry){
+void cat_str::output(map<String, String, less<String> > &menuentry){
   genoutput(soutput(menuentry),menuentry);
 }
 
 ostream &var_str::output(ostream &o,
-			 map<string, string, less<string> > &menuentry){
+			 map<String, String, less<String> > &menuentry){
   return o<<menuentry[var_name];
 }
 ostream &func_str::output(ostream &o,
-			 map<string, string, less<string> > &menuentry){
+			 map<String, String, less<String> > &menuentry){
   return f->output(o,args,menuentry);
 }
 
 /////////////////////////////////////////////////////
 //  Debug routines
 //
-ostream &const_str::debug(ostream &o){
+ostream &const_str::debuginfo(ostream &o){
   return o<<"CONST_STR: "<<data<<endl;
 }
-ostream &cat_str::debug(ostream &o){
+ostream &cat_str::debuginfo(ostream &o){
   vector<str *>::iterator i;
   o<<"CAT_STR: "<<endl;
   for(i=v.begin();i!=v.end();i++){
-    (*i)->debug(o);
+    (*i)->debuginfo(o);
   }
   return o;
 }
-ostream &var_str::debug(ostream &o){
+ostream &var_str::debuginfo(ostream &o){
   return o<<"VAR_STR: "<<var_name<<endl;
 }
-ostream &func_str::debug(ostream &o){
+ostream &func_str::debuginfo(ostream &o){
   o<<"FUNC_STR: "<<f->name()<<" ("<<endl;
   vector<cat_str *>::iterator i;  
   for(i=args.begin();i!=args.end();i++){
     o<<", ";
-    (*i)->debug(o);
+    (*i)->debuginfo(o);
   }
   return o<<")"<<endl;
 }
@@ -285,12 +291,12 @@ ostream &func_str::debug(ostream &o){
 //
 
 ostream &prefix_func::output(ostream &o, vector<cat_str *> &,
-			     map<string, string, less<string> > &){
+			     map<String, String, less<String> > &){
 
   return o<<config->prefix();
 }
 ostream &ifroot_func::output(ostream &o, vector<cat_str *> & args,
-			     map<string, string, less<string> > &menuentry){
+			     map<String, String, less<String> > &menuentry){
   if(getuid())
     args[1]->output(o,menuentry);
   else
@@ -300,35 +306,35 @@ ostream &ifroot_func::output(ostream &o, vector<cat_str *> & args,
 }
 
 ostream &print_func::output(ostream &o, vector<cat_str *> &args,
-			    map<string, string, less<string> > &menuentry){
+			    map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
 
-  if(empty_string(s)){
+  if(empty_String(s)){
     cerr<<"Zero-size argument to print function";
     throw informed_fatal();
   }
   return o<<s;
 }
 ostream &ifempty_func::output(ostream &o, vector<cat_str *> &args,
-			      map<string, string, less<string> > &menuentry){
+			      map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
 
-  if(empty_string(s))
+  if(empty_String(s))
     args[1]->output(o,menuentry);
   return o;
 }
 ostream &ifnempty_func::output(ostream &o, vector<cat_str *> &args,
-			       map<string, string, less<string> > &menuentry){
+			       map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
-  if(!empty_string(s))
+  if(!empty_String(s))
     args[1]->output(o,menuentry);
   return o;
 }
 ostream &iffile_func::output(ostream &o, vector<cat_str *> &args,
-			     map<string, string, less<string> > &menuentry){
+			     map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   ifstream f(s);
@@ -336,22 +342,44 @@ ostream &iffile_func::output(ostream &o, vector<cat_str *> &args,
     args[1]->output(o,menuentry);
   return o;
 }
+ostream &ifelsefile_func::output(ostream &o, vector<cat_str *> &args,
+			     map<String, String, less<String> > &menuentry){
+
+  String s=args[0]->soutput(menuentry);
+  ifstream f(s);
+  if(f)
+    args[1]->output(o,menuentry);
+  else
+    args[2]->output(o,menuentry);
+  return o;
+}
+ostream &catfile_func::output(ostream &o, vector<cat_str *> &args,
+			     map<String, String, less<String> > &menuentry){
+
+  String s=args[0]->soutput(menuentry);
+  ifstream f(s);
+  char c;
+
+  while(f && f.get(c))
+    o<<c;
+  return o;
+}
 
 ostream &esc_func::output(ostream &o, vector<cat_str *> &args,
-			  map<string, string, less<string> > &menuentry){
+			  map<String, String, less<String> > &menuentry){
 
-  return o<<escape_string(args[0]->soutput(menuentry),
+  return o<<escape_String(args[0]->soutput(menuentry),
 			  args[1]->soutput(menuentry));
 }
 ostream &escwith_func::output(ostream &o, vector<cat_str *> &args,
-			      map<string, string, less<string> > &menuentry){
+			      map<String, String, less<String> > &menuentry){
 
-  return o<<escapewith_string(args[0]->soutput(menuentry),
+  return o<<escapewith_String(args[0]->soutput(menuentry),
 			      args[1]->soutput(menuentry),
 			      args[2]->soutput(menuentry));
 }
 ostream &escfirst_func::output(ostream &o, vector<cat_str *> &args,
-			      map<string, string, less<string> > &menuentry){
+			      map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   String esc=args[1]->soutput(menuentry);
@@ -368,32 +396,78 @@ ostream &escfirst_func::output(ostream &o, vector<cat_str *> &args,
   }
   return o<<t;
 }
-ostream &replacewith_func::output(ostream &o, vector<cat_str *> &args,
-				  map<string, string, less<string> > &menuentry){
+ostream &tolower_func::output(ostream &o, vector<cat_str *> &args,
+			      map<String, String, less<String> > &menuentry){
 
-  return o<<replacewith_string(args[0]->soutput(menuentry),
+  return o<<tolower_String(args[0]->soutput(menuentry));
+}
+ostream &toupper_func::output(ostream &o, vector<cat_str *> &args,
+				  map<String, String, less<String> > &menuentry){
+
+  return o<<toupper_String(args[0]->soutput(menuentry));
+}
+ostream &replacewith_func::output(ostream &o, vector<cat_str *> &args,
+				  map<String, String, less<String> > &menuentry){
+
+  return o<<replacewith_String(args[0]->soutput(menuentry),
 			       args[1]->soutput(menuentry),
 			       args[2]->soutput(menuentry));
 }
 ostream &cppesc_func::output(ostream &o, vector<cat_str *> &args,
-			      map<string, string, less<string> > &menuentry){
+			      map<String, String, less<String> > &menuentry){
 
-  return o<<cppesc_string(args[0]->soutput(menuentry));
+  return o<<cppesc_String(args[0]->soutput(menuentry));
+}
+
+ostream &add_func::output(ostream &o, vector<cat_str *> &args,
+			  map<String, String, less<String> > &menuentry){
+
+  int x=Stringtoi(args[0]->soutput(menuentry));
+  int y=Stringtoi(args[1]->soutput(menuentry));
+
+  return o<<itoString(x+y);
+}
+ostream &sub_func::output(ostream &o, vector<cat_str *> &args,
+			  map<String, String, less<String> > &menuentry){
+
+  int x=Stringtoi(args[0]->soutput(menuentry));
+  int y=Stringtoi(args[1]->soutput(menuentry));
+
+  return o<<itoString(x-y);
+}
+ostream &mult_func::output(ostream &o, vector<cat_str *> &args,
+			  map<String, String, less<String> > &menuentry){
+
+  int x=Stringtoi(args[0]->soutput(menuentry));
+  int y=Stringtoi(args[1]->soutput(menuentry));
+
+  return o<<itoString(x*y);
+}
+ostream &div_func::output(ostream &o, vector<cat_str *> &args,
+			  map<String, String, less<String> > &menuentry){
+
+  int x=Stringtoi(args[0]->soutput(menuentry));
+  int y=Stringtoi(args[1]->soutput(menuentry));
+  
+  if(y)
+    return o<<itoString(x/y);
+  else
+    return o<<"0";
 }
 
 ostream &ifelse_func::output(ostream &o, vector<cat_str *> &args,
-			     map<string, string, less<string> > &menuentry){
+			     map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
 
-  (!empty_string(s)) ?
+  (!empty_String(s)) ?
     args[1]->output(o,menuentry)
     :
     args[2]->output(o,menuentry);
   return o;
 }
 ostream &ifeq_func::output(ostream &o, vector<cat_str *> &args,
-			   map<string, string, less<string> > &menuentry){
+			   map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   String t=args[1]->soutput(menuentry);
@@ -403,7 +477,7 @@ ostream &ifeq_func::output(ostream &o, vector<cat_str *> &args,
   return o;
 }
 ostream &ifneq_func::output(ostream &o, vector<cat_str *> &args,
-			   map<string, string, less<string> > &menuentry){
+			   map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   String t=args[1]->soutput(menuentry);
@@ -413,7 +487,7 @@ ostream &ifneq_func::output(ostream &o, vector<cat_str *> &args,
   return o;
 }
 ostream &ifeqelse_func::output(ostream &o, vector<cat_str *> &args,
-			   map<string, string, less<string> > &menuentry){
+			   map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   String t=args[1]->soutput(menuentry);
@@ -426,11 +500,11 @@ ostream &ifeqelse_func::output(ostream &o, vector<cat_str *> &args,
 }
 
 ostream &cond_surr_func::output(ostream &o, vector<cat_str *> &args,
-				map<string, string, less<string> > &menuentry){
+				map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
 
-  if(!empty_string(s)){
+  if(!empty_String(s)){
     args[1]->output(o,menuentry);
     args[0]->output(o,menuentry);
     args[2]->output(o,menuentry);
@@ -439,7 +513,7 @@ ostream &cond_surr_func::output(ostream &o, vector<cat_str *> &args,
 }
 
 ostream &parent_func::output(ostream &o, vector<cat_str *> &args,
-			     map<string, string, less<string> > &menuentry){
+			     map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
 
@@ -447,7 +521,7 @@ ostream &parent_func::output(ostream &o, vector<cat_str *> &args,
 }
 
 ostream &basename_func::output(ostream &o, vector<cat_str *> &args,
-			       map<string, string, less<string> > &menuentry){
+			       map<String, String, less<String> > &menuentry){
 
   String s=args[0]->soutput(menuentry);
   String t;
@@ -463,64 +537,89 @@ ostream &basename_func::output(ostream &o, vector<cat_str *> &args,
   else
     return o<<s.at(q+1,p-q-1);
 }
+ostream &entrycount_func::output(ostream &o, vector<cat_str *> &/*args*/,
+				 map<String, String, less<String> > &menuentry){
+  return o<<menuentry[PRIVATE_ENTRYCOUNT_VAR];
+}
+ostream &entryindex_func::output(ostream &o, vector<cat_str *> &/*args*/,
+				 map<String, String, less<String> > &menuentry){
+  return o<<menuentry[PRIVATE_ENTRYINDEX_VAR];
+}
+
+/////////////////////////////////////////////////////
+//  "defined" function (macro).
+//
+
+ostream &func_def::output(ostream &o, vector<cat_str *> &args,
+			  map<String, String, less<String> > &menuentry){
+
+  String t;
+  unsigned int i;
+  map<String, String, less<String> > local_menuentry=menuentry;
+
+  for(i=0; i<args_name.size(); i++)
+    local_menuentry[args_name[i]]=args[i]->soutput(menuentry);
+
+  f->output(o,local_menuentry);
+  return o;
+}
+func_def::func_def(parsestream &i){
+  char c;
+
+  n=i.get_name();
+  i.skip_space();
+  i.skip_char('(');
+  do{
+    i.skip_space();
+    c=i.get_char();
+    if(c==')')
+      break;
+    i.put_back(c);
+    i.skip_char('$');
+    args_name.push_back(i.get_name());
+    i.skip_space();
+  } while((c=i.get_char())&&(c==','));
+  i.put_back(c);
+  i.skip_char(')');
+  i.skip_space();
+  i.skip_char('=');
+  f=new cat_str(i);
+}
 
 /////////////////////////////////////////////////////
 //  Supported stuff
 //
 
 supportedinfo::supportedinfo(parsestream &i){
-  char c=-1;
   linenumber=1;
   int prec=0;
-  try{
-    while(1){
-      string name;
-      while(1){
-	i.skip_space();
-	c=i.get_char();
-	if(!c)
-	  break;
-	if(c=='#'){
-	  i.skip_line();
-	  break;
-	}
-	i.put_back(c);
-	name=i.get_name();
-	if(name=="endsupported"){
-	  i.skip_line();
-	  return;
-	}
-	name=upcase(name);
-	if(dodebug)
-	  cerr<<"SUPPORTED_CONSTRUCT: name="<<name<<endl;
-	i.skip_space();
-	i.skip_char('=');
-	
-	supinf inf;
-	inf.c=new cat_str(i);
-	inf.prec=prec++;
-	sup[name]=inf;
-	if(dodebug)
-	  sup[name].c->debug(cerr);
-      }
+  while(1){
+    try{
+      String name;
+      i.skip_space();
+      name=i.get_name();
+      if(name=="endsupported")
+	return;
+      name=upcase(name);
+      if(dodebug)
+	cerr<<"SUPPORTED_CONSTRUCT: name="<<name<<endl;
+      
+      supinf inf;
+      inf.c=get_eq_cat_str(i);
+      inf.prec=prec++;
+      sup[name]=inf;
+      if(dodebug)
+	sup[name].c->debuginfo(cerr);
       i.skip_line();   //read away the final newline
-    }  
-  }
-  catch(endoffile){}
-  catch(char_expected c){
-    cerr<<"\""<<c.c<<" expected in line "<<i.linenumber()<<", while parsing supported stuff"<<endl;
-    throw informed_fatal();
-  }
-  catch(char_unexpected c){
-    cerr<<"unexpected character \""<<c.c<<"\" at line "<<i.linenumber()<<", while parsing supported stuff"<<endl;
-    throw informed_fatal();
+    }
+    catch(endofline d){}
   }
 }
 
-void supportedinfo::subst(map<string, string, less<string> > vars){
+void supportedinfo::subst(map<String, String, less<String> > vars){
 
-  map<string, string, less<string> >::iterator i;
-  map<string, supinf, less<string> >::iterator j;
+  map<String, String, less<String> >::iterator i;
+  map<String, supinf, less<String> >::iterator j;
 
   if((i=vars.find(NEEDS_VAR))==vars.end()){
     cerr<<"Undefined "NEEDS_VAR" variable in menuentries"<<endl;
@@ -532,8 +631,8 @@ void supportedinfo::subst(map<string, string, less<string> > vars){
   }
   genoutput((*j).second.c->soutput(vars), vars);
 }
-int supportedinfo::prec(string &name){
-  map<string, supinf, less<string> >::iterator i;
+int supportedinfo::prec(String &name){
+  map<String, supinf, less<String> >::iterator i;
   int p;
   if((i=sup.find(upcase(name)))==sup.end())
     p=MAXINT;
@@ -545,11 +644,11 @@ int supportedinfo::prec(string &name){
 }
 
 ostream &supportedinfo::debuginfo(ostream &o){
-  map<string, supinf, less<string> >::iterator i;
+  map<String, supinf, less<String> >::iterator i;
   for(i=sup.begin();i!=sup.end();i++){
     o<<"SUPPORTED:** name="<<(*i).first<<", prec="
 	<<(*i).second.prec<<" Def="<<endl;
-    (*i).second.c->debug(o);
+    (*i).second.c->debuginfo(o);
   }
   return o;
 }
@@ -557,8 +656,7 @@ ostream &supportedinfo::debuginfo(ostream &o){
 //  configinfo
 //
 configinfo::configinfo(parsestream &i){
-  char c;
-  string errmsg("Parse error: String constant expected");
+  String errmsg("Parse error: String constant expected");
   linenumber=1;
   treew="c(m)";
   prerun=postrun=genmenu=hkexclude=startmenu=endmenu=submenutitle=NULL;
@@ -566,19 +664,14 @@ configinfo::configinfo(parsestream &i){
   roots="/Debian";
   try{
     while(1){
-      while((c=i.get_char())){
-	string name,arg;
-	if(c=='#'){
-	  i.skip_line();
-	  continue;
-	} else
-	  i.put_back(c);
-	name=i.get_name();
-	i.skip_space();
+        String name;
+        name=i.get_name();
 	if(name=="supported"){
 	  i.skip_line();
 	  supported=new supportedinfo(i);
 	}
+	else if(name=="function")
+	  store_func(new func_def(i));
 	else if(name=="startmenu")
 	  startmenu=get_eq_cat_str(i);
 	else if(name=="endmenu")
@@ -595,35 +688,35 @@ configinfo::configinfo(parsestream &i){
 	  prerun=get_eq_cat_str(i);	
 
 	else if(name=="compat"){
-	  if((compt=i.get_eq_stringconst())!="menu-1"){
+	  if((compt=i.get_eq_Stringconst())!="menu-1"){
 	    cerr<<"Unknown compat mode "<<compt<<" in config file (only know of \""COMPAT_MODE"\"";
 	    throw informed_fatal();
 	  }
 	}
 	else if(name=="rcfile")
-	  rcf=i.get_eq_stringconst();
+	  rcf=i.get_eq_Stringconst();
 	else if(name=="examplercfile")
-	  exrcf=i.get_eq_stringconst();
+	  exrcf=i.get_eq_Stringconst();
 	else if(name=="mainmenutitle")
-	  mainmt=i.get_eq_stringconst();
+	  mainmt=i.get_eq_Stringconst();
 	else if(name=="rootsection")
-	  roots=i.get_eq_stringconst();
+	  roots=i.get_eq_Stringconst();
 	else if(name=="rootprefix")
-	  rootpref=i.get_eq_stringconst();
+	  rootpref=i.get_eq_Stringconst();
 	else if(name=="userprefix")
-	  userpref=i.get_eq_stringconst();
+	  userpref=i.get_eq_Stringconst();
 	else if(name=="treewalk")
-	  treew=i.get_eq_stringconst();
+	  treew=i.get_eq_Stringconst();
 	else if(name=="postoutput")
-	  postout=i.get_eq_stringconst();
+	  postout=i.get_eq_Stringconst();
 	else if(name=="preoutput")
-	  preout=i.get_eq_stringconst();
+	  preout=i.get_eq_Stringconst();
 	else if(name=="command"){
-	  system(i.get_eq_stringconst());
+	  system(i.get_eq_Stringconst());
 	  exit(0);
 	}
 	else if(name=="hotkeycase"){
-	  String s=i.get_eq_stringconst();
+	  String s=i.get_eq_Stringconst();
 	  if(s=="sensitive")
 	    hotkeycase=1;
 	  else if (s=="insensitive")
@@ -633,34 +726,23 @@ configinfo::configinfo(parsestream &i){
 	    throw informed_fatal();
 	  }
 	}
-	else{
-	  cerr<<"Unknown identifier \""<<name<<"\" in script"<<endl;
-	  throw informed_fatal();
-	}
-      }
+	else
+	  throw unknown_ident(&i);
       i.skip_line();//read away final newline
     }
   }
   catch(endoffile){}
-  catch(char_expected c){
-    cerr<<"\""<<c.c<<" expected in line "<<i.linenumber()<<", while parsing startupscript"<<endl;
+  
+  check_config();
+}
+void configinfo::check_config(){
+  if(!(genmenu && startmenu && endmenu)){
+    cerr<<"At least one of genmenu, startmenu, endmenu"<<endl
+	<<"is undefined in the config file. All of these have to be "<<endl
+	<<"defined (although they may be equal to \"\")"<<endl;
     throw informed_fatal();
   }
-  catch(char_unexpected c){
-    cerr<<"unexpected character \""<<c.c<<"\" at line "<<i.linenumber()<<", while parsing startupscript"<<endl;
-    throw informed_fatal();
-  }
-  catch(ident_expected){
-    cerr<<"Identifier expected in line "<<i.linenumber()<<", while parsing startupscript"<<endl;
-    throw informed_fatal();
-  }
-  catch(narg_mismatch f){
-    cerr<<"Number of arguments to function "<<f.name<<" doesn't match, at line "<<i.linenumber()<<endl;
-    throw informed_fatal();
-  }
-
 }
-
 String configinfo::prefix(){
   if(getuid())
     return String(getenv("HOME"))+"/"+userprefix();
@@ -668,7 +750,7 @@ String configinfo::prefix(){
     return rootpref;
 }
 
-ostream &configinfo::debug(ostream &o){
+ostream &configinfo::debuginfo(ostream &o){
     o<<"Using compatibility with:"<<compt<<endl
      <<"mainmenutitle   : "<<mainmt   <<endl
      <<"rootsection     : "<<roots    <<endl
@@ -678,16 +760,16 @@ ostream &configinfo::debug(ostream &o){
      <<"user-prefix     : "<<userpref <<endl
      <<"startmenu       : "<<endl;
     if(startmenu)
-      startmenu->debug(o);
+      startmenu->debuginfo(o);
     o<<"endmenu         : "<<endl;
     if(endmenu)
-      endmenu->debug(o);
+      endmenu->debuginfo(o);
     o<<"genmenu         : "<<endl;
     if(genmenu)
-      genmenu->debug(o);
+      genmenu->debuginfo(o);
     o<<"submenutitle    : "<<endl;
     if(submenutitle)
-      submenutitle->debug(o);
+      submenutitle->debuginfo(o);
   o<<"mainmenutitle   : "<<mainmt   <<endl
    <<"treewalk        : "<<treew    <<endl;
   return o;
@@ -697,13 +779,13 @@ ostream &configinfo::debug(ostream &o){
 //  Menuentry
 //
 
-void menuentry::add_entry(string sec, string fullsec, 
-			  map<string, string, less<string> > &v){
+void menuentry::add_entry(String sec, String fullsec, 
+			  map<String, String, less<String> > &v){
   if(dodebug)
     cerr<<"ADDENTRY: sec="<<sec<<"needs="<<v["needs"]<<endl;
   if(sec.length()!=0){
-    string name;
-    map <string, menuentry *, less<string> >::iterator i;
+    String name;
+    map <String, menuentry *, less<String> >::iterator i;
     if(sec.contains("/")){
       name=sec.before("/");
       sec=sec.after("/");
@@ -739,8 +821,8 @@ void menuentry::add_entry(string sec, string fullsec,
   vars[SECTION_VAR]=fullsec;
 }
 void menuentry::output(){
-  string treew=config->treewalk();
-  map <string, menuentry *, less<string> >::iterator i;
+  String treew=config->treewalk();
+  map <String, menuentry *, less<String> >::iterator i;
 
   for(unsigned int j=0;j<treew.length(); j++){
     bool children_too=false;
@@ -777,11 +859,19 @@ void menuentry::output(){
 }
 
 void menuentry::postprocess(){
-  map <string, menuentry *, less<string> >::iterator i;
-  for(i=submenus.begin(); i!=submenus.end(); i++)
+  map <String, menuentry *, less<String> >::iterator i;
+  int index;
+  for(i=submenus.begin(), index=0; 
+      i!=submenus.end(); 
+      i++, index++){
+    (*i).second->vars[PRIVATE_ENTRYINDEX_VAR]=itoString(index);
     if((*i).second->submenus.size())
       (*i).second->postprocess();
+    else
+      (*i).second->vars[PRIVATE_ENTRYCOUNT_VAR]="0";
+  }
   generate_hotkeys();
+  vars[PRIVATE_ENTRYCOUNT_VAR]=itoString(submenus.size());
 }
 char menuentry::hotkeyconv(char h){
   if (config->hotkeycase)
@@ -792,8 +882,8 @@ char menuentry::hotkeyconv(char h){
 void menuentry::generate_hotkeys(){
   unsigned int i,j;
   list<int>::iterator k, old_k;
-  map <string, menuentry *, less<string> >::iterator subi;
-  map <string, string, less<string> >::iterator l;
+  map <String, menuentry *, less<String> >::iterator subi;
+  map <String, String, less<String> >::iterator l;
   vector<String> keys;
   list<int> todo;
   set<char, less<char> > used_chars;
@@ -849,25 +939,43 @@ void menuentry::generate_hotkeys(){
 //  Misc
 //
 
+map <String, String, less<String> > read_vars(parsestream &i){
+  map <String, String, less<String> > m;
+  try{
+    String name, val;
+    while(1){
+      name=i.get_name();
+      val=i.get_eq_Stringconst();
+      m[name]=val;
+    } 
+  }
+  catch(endofline p){};
+  return m;
+}
+void check_vars(parsestream &i,
+		map <String, String, less<String> > &m){
+  vector <String> need;
+  map <String, String, less<String> >::iterator j;
+  unsigned int k;
+
+  need.push_back(SECTION_VAR);
+  need.push_back(TITLE_VAR);
+  need.push_back(NEEDS_VAR);
+
+  for(k=0;k<need.size();k++){
+    j=m.find(need[k]);
+    if((j==m.end())||((*j).second==""))
+      throw missing_tag(&i,need[k]);
+  }
+}
 void read_input(parsestream &i){
-  char c;
   String s;
   try{
     while(1){
-      map <string, string, less<string> > m;
-      string name, val;
-      string sec;
-      do{
-	name=i.get_name();
-	val=i.get_eq_stringconst();
-	m[name]=val;
-	try{
-	  c=i.get_char();
-	  i.put_back(c);
-	}catch(endoffile){
-	  c=0;
-	}
-      } while(c);
+      map <String, String, less<String> > m;
+      String sec;
+      m=read_vars(i);
+      check_vars(i,m);
       sec=m[SECTION_VAR]+'/';
       if(m.find(SORT_VAR)!=m.end())
 	sec=sec+m[SORT_VAR]+":";
@@ -877,24 +985,12 @@ void read_input(parsestream &i){
       i.skip_line();   //read away the final newline
     }
   }
-  catch(endoffile){}
-  catch(char_expected c){
-    cerr<<"\""<<c.c<<" expected in line "<<i.linenumber()<<", while parsing stdinput"<<endl;
-    throw informed_fatal();
-  }
-  catch(char_unexpected c){
-    cerr<<"unexpected character \""<<c.c<<"\" at line "<<i.linenumber()<<", while parsing stdinput"<<endl;
-    throw informed_fatal();
-  }
-  catch(ident_expected){
-    cerr<<"Identifier expected in line "<<i.linenumber()<<", while parsing stdinput"<<endl;
-    throw informed_fatal();
-  }
-
+  catch(endoffile p){}
 }
 
-void includemenus(string o, string i,
-                  string rep, string m){
+
+void includemenus(String o, String i,
+                  String rep, String m){
   //copy filename i to filename o, replacing the line
   //rep with menu-file m
   char buf[MAX_BUF];
@@ -902,12 +998,14 @@ void includemenus(string o, string i,
   bool changed=false;
   ifstream fi(i);
   ifstream fm(m);
-  ofstream fo(o);
   
   if(!fi){
     cerr<<"Cannot open file "<<i<<endl; throw informed_fatal();}
   if(!fm){
     cerr<<"Cannot open file "<<m<<endl; throw informed_fatal();}
+
+  ofstream fo(o);
+
   if(!fo){
     cerr<<"Cannot open file "<<o<<endl; throw informed_fatal();}
 
@@ -930,12 +1028,11 @@ void includemenus(string o, string i,
 	<<"\" didn't occur in example file "<<i<<endl;
 }
 
-
 int main(int argc, char **argv){
   int c, option_index;
   char script[MAX_BUF];
-  //  ofstream *menudef;
-  string menudefname;
+
+  String menudefname;
   
   try{
     if(argv[1])
@@ -959,31 +1056,26 @@ int main(int argc, char **argv){
       case 'f':break;
       }
     } while (c!=-1);
-    ifstream i(script);
-    if(!i){
+
+    add_functions();
+    parsestream ps(script);
+    if(!ps.good()){
       cerr<<"Cannot open script "<<script<< " for reading"<<endl;
       throw informed_fatal();
     }
-    parsestream ps(i);
     config=new configinfo(ps);
-    if(dodebug)
+    if(dodebug){
+      config->debuginfo(cerr);
       supported->debuginfo(cerr);
+    }
     if(config->prerun)
       system(config->prerun->soutput(menu.vars));
-    /*    menudefname=userprefix+config->genmenu();
-	  menudef=new ofstream(menudefname);
-	  if(!(*menudef)){
-	  cerr<<"Cannot open "<<menudefname<<" for writing";
-	  throw informed_fatal();
-	  }
-    */
     parsestream psscript(cin);
     read_input(psscript);
     if(menu.vars[TITLE_VAR]=="")
       menu.vars[TITLE_VAR]=config->mainmenutitle();
     menu.postprocess();
     menu.output();
-    //    delete menudef;
     closegenoutput();
     if(config->rcfile().length()&&config->examplercfile().length())
       includemenus(config->prefix()+"/"+config->rcfile(),
@@ -992,13 +1084,29 @@ int main(int argc, char **argv){
 		   config->prefix()+"/"+config->genmenu->soutput(menu.vars));
     if(config->postrun)
       system(config->postrun->soutput(menu.vars));
+    exit(0);
   }
-  catch(informed_fatal){
-    cerr<<argv[1]<<": Aborting."<<endl;
-    exit(1);
-  }
-  catch(dir_createerror d){
-    cerr<<argv[1]<<": Cannot create directory "<<d.name<<endl;
-  }
-}
+  //yes, I know this _should_ be handled in one
+  //"derived" exception handler, but gcc doesn't like those
+  // (I'm getting internal compiler errors now quite often
+  // already, and I remember having them even more frequenly
+  // with the -frtty stuff -- _I_WANT_gcc-2.8_!).
+  catch(endoffile p){        p.report(cerr); }
+  catch(endofline p){        p.report(cerr); }
+  catch(char_expected p){    p.report(cerr); }
+  catch(char_unexpected p){  p.report(cerr); }
+  catch(ident_expected p){   p.report(cerr); }
+  catch(narg_mismatch p){    p.report(cerr); }
+  catch(unknown_function p){ p.report(cerr); }
+  catch(unknown_ident p){    p.report(cerr); }
+  catch(ferror_open p){      p.report(cerr); }
+  catch(missing_tag p){      p.report(cerr); }
+  //catch(dir_createerror d){
+  //  cerr<<": Cannot create directory "<<d.name<<endl;
+  //}
+  catch(informed_fatal){};
 
+  cerr<<script<<": Aborting"<<endl;
+
+  exit(1);
+}
