@@ -1,4 +1,4 @@
-/* Copyright: GPL */
+/* Copyright:  GPL */
 /*
   Witten by joost witteveen;  
   read_pkginfo function by Tom Lees, run_menumethods by both.
@@ -17,11 +17,15 @@
 #include <ctype.h>
 #include <dirent.h>
 #include <signal.h>
+#include <syslog.h>
+#include <errno.h>
+#include <procbuf.h>
 
 int debug=0, verbose=0;
 set <String, less<String> > installed_packages;
 set <String, less<String> > menufiles_processed;
 translateinfo *transinfo;
+configinfo     config;
 
 DIR *open_dir_check(String dirname){
   struct stat st;
@@ -32,14 +36,40 @@ DIR *open_dir_check(String dirname){
 
   return opendir (&(dirname[0]));
 }
-
+bool executable(const char *s){
+  struct stat st;
+  stat (s, &st);
+  return (((st.st_mode & S_IXOTH) || 
+	   ((st.st_mode & S_IXGRP) && st.st_gid == getegid ()) || 
+	   ((st.st_mode & S_IXUSR) && st.st_uid == geteuid ())) && 
+	  (S_ISREG(st.st_mode) || S_ISLNK(st.st_mode)));
+}
 
 /////////////////////////////////////////////////////
 //  menuentry:
 //
 
 bool menuentry::test_installed(String filename){
-  return installed_packages.find(filename)!=installed_packages.end();
+  if(filename.contains("local.",0))
+    return true;
+  else
+    return installed_packages.find(filename)!=installed_packages.end();
+}
+
+bool menuentry::check_install(parsestream &i){
+  String name, val;
+
+  name=i.get_name();
+  if(name!=COND_PACKAGE)
+    throw unknown_cond_package(&i);
+  i.skip_char('(');
+  val=i.get_name();
+  if(!test_installed(val)){
+    i.skip_line();
+    throw cond_inst_false();
+  }
+  i.skip_char(')');
+  return true;
 }
 menuentry::menuentry(parsestream &i, const String &filename){
   char c;
@@ -49,41 +79,33 @@ menuentry::menuentry(parsestream &i, const String &filename){
     //new format menuentry
     
     //read available info
-    name=i.get_name();
-    if(name!=COND_PACKAGE)
-      throw unknown_cond_package();
-    i.skip_char('(');
-    val=i.get_name();
-    if(!test_installed(val))
-      throw cond_inst_false();
-    i.skip_char(')');
+    check_install(i);
     i.skip_space();
     i.skip_char(':');
     //read menuentry:
     try{
       do{
 	name=i.get_name();
-	val=i.get_eq_stringconst();
+	val=i.get_eq_Stringconst();
 	data[name]=val;
 	c=i.get_char();
 	i.put_back(c);
       }while(c);
     }
-    catch(endoffile){};
+    catch(endofline){};
     i.skip_line();
   }else{
     //old format menuentry
-    if((!test_installed(filename))
-       &&(!filename.contains("local",0))){
+    if(!test_installed(filename)){
       i.skip_line();
       throw cond_inst_false();
     }
     i.put_back(c);
-    data[NEEDS_VAR]  =i.get_stringconst();
-    data[SECTION_VAR]=i.get_stringconst();
-    i.get_stringconst(); //id is unused
-    data[ICON_VAR]   =i.get_stringconst();
-    data[TITLE_VAR]  =i.get_stringconst();
+    data[NEEDS_VAR]  =i.get_Stringconst();
+    data[SECTION_VAR]=i.get_Stringconst();
+    i.get_Stringconst(); //id is unused
+    data[ICON_VAR]   =i.get_Stringconst();
+    data[TITLE_VAR]  =i.get_Stringconst();
     i.skip_space();
     data[COMMAND_VAR]=i.get_line();
   }
@@ -95,7 +117,7 @@ void menuentry::output(vector<String> &s){
   if((i=data.begin())!=data.end())
     while (true){
       t+=String((*i).first) + "=\"" + 
-	 escape_string((*i).second,"\"\n") + "\"";
+	 escape_String((*i).second,"\"\n") + "\"";
       i++;
       if(i==data.end()){
 	break;
@@ -103,25 +125,21 @@ void menuentry::output(vector<String> &s){
 	t+=" ";
     }
   t+="\n";
-  if(debug)
-    cerr<<"ADDING:"<<t;
+  config.report(String("ADDING")+t,configinfo::report_debug);
   s.push_back(t);
 }
 void menuentry::output_compat(vector<String> &s){
   String t,icon;
-  static idcount=0;
   icon=data[ICON_VAR];
   if(icon=="")
     icon="none";
   t=data[NEEDS_VAR]  +" "
    +data[SECTION_VAR]+" id"
-   +itoString(idcount++)+" "  
+   +cppesc_String(data[TITLE_VAR])+" "  
    +icon   +" \""
    +data[TITLE_VAR]  +"\" "
-   +escape_string(data[COMMAND_VAR],"\n")+"\n";
-  if(debug)
-    cerr<<"ADDING:"<<t;
-
+   +escape_String(data[COMMAND_VAR],"\n")+"\n";
+  config.report(String("ADDING")+t,configinfo::report_debug);
   s.push_back(t);
 }
 ostream &menuentry::debugoutput(ostream &o){
@@ -133,6 +151,97 @@ ostream &menuentry::debugoutput(ostream &o){
   return o;
 }
 
+/////////////////////////////////////////////////////
+//  configinfo
+//
+configinfo::configinfo(){
+  verbosity=report_normal;
+  method=method_stderr;
+}
+void configinfo::parse_def(const String &var, parsestream &p){
+  String s;
+  s=p.get_Stringconst();
+  if(var=="verbosity"){
+    if     (s=="quiet")   verbosity=report_quiet;
+    else if(s=="normal")  verbosity=report_normal;
+    else if(s=="verbose") verbosity=report_verbose;
+    else if(s=="debug")   verbosity=report_debug;
+    else throw def_error();
+  } else if(var=="method"){
+    if     (s=="stdout")  method=method_stdout;
+    else if(s=="stderr")  method=method_stderr;
+    else if(s=="syslog") { 
+      method=method_syslog;
+      String facility;
+      String priority;
+      facility=p.get_Stringconst();
+      p.skip_char('.');
+      priority=p.get_Stringconst();
+           if(facility=="auth")       syslog_facility=LOG_AUTH;
+      else if(facility=="authpriv")   syslog_facility=LOG_AUTHPRIV;
+      else if(facility=="authcron")   syslog_facility=LOG_CRON;
+      else if(facility=="authdaemon") syslog_facility=LOG_AUTHPRIV;
+      else if(facility=="authkern")   syslog_facility=LOG_KERN;
+      else if(facility=="authlocal0") syslog_facility=LOG_LOCAL0;
+      else if(facility=="authlocal1") syslog_facility=LOG_LOCAL1;
+      else if(facility=="authlocal2") syslog_facility=LOG_LOCAL2;
+      else if(facility=="authlocal3") syslog_facility=LOG_LOCAL3;
+      else if(facility=="authlocal4") syslog_facility=LOG_LOCAL4;
+      else if(facility=="authlocal5") syslog_facility=LOG_LOCAL5;
+      else if(facility=="authlocal6") syslog_facility=LOG_LOCAL6;
+      else if(facility=="authlocal7") syslog_facility=LOG_LOCAL7;
+      else if(facility=="authlpr")    syslog_facility=LOG_LPR;
+      else if(facility=="authmail")   syslog_facility=LOG_MAIL;
+      else if(facility=="authnews")   syslog_facility=LOG_NEWS;
+      else if(facility=="authsyslog") syslog_facility=LOG_SYSLOG;
+      else if(facility=="authuser")   syslog_facility=LOG_USER;
+      else if(facility=="authuucp")   syslog_facility=LOG_UUCP;
+      else throw def_error();
+
+	   if(priority=="emerg")      syslog_priority=LOG_EMERG;
+      else if(priority=="alert")      syslog_priority=LOG_ALERT;
+      else if(priority=="crit")       syslog_priority=LOG_CRIT;
+      else if(priority=="err")        syslog_priority=LOG_ERR;
+      else if(priority=="warning")    syslog_priority=LOG_WARNING;
+      else if(priority=="notice")     syslog_priority=LOG_NOTICE;
+      else if(priority=="info")       syslog_priority=LOG_INFO;
+      else if(priority=="debug")      syslog_priority=LOG_DEBUG;
+      else throw def_error();
+    }
+    else throw def_error();
+  }
+  else throw def_error();
+}
+
+void configinfo::update(String filename){
+  parsestream p(filename);
+  try{
+    while(true){
+      String var;
+      var=p.get_name();
+      p.skip_char('=');
+      parse_def(var,p);
+      p.skip_line();
+    }
+  } catch (endoffile){};
+}
+void configinfo::report(const String &message,
+			verbosity_type v){
+  if(v<=verbosity){
+    switch(method){
+    case method_stdout:
+      cout<<message<<endl;
+      break;
+    case method_stderr:
+      cerr<<message<<endl;
+      break;
+    case method_syslog:
+      openlog("update-menus",LOG_PID,syslog_facility);
+      syslog(syslog_priority,&(message[0]));
+      closelog();
+    }
+  }
+}
 /////////////////////////////////////////////////////
 //  translate stuff
 //
@@ -146,8 +255,7 @@ trans_class::trans_class(const String &m,
 }
 
 bool  trans_class::check(String &s){
-  if(debug)
-    cerr<<"checking "<<match<<" < "<<s<<endl;
+  config.report(String("checking ")+match+" < "+s,configinfo::report_debug);
   return match.contains(s,0);
 }
 
@@ -182,21 +290,16 @@ void translateinfo::init(parsestream &i){
   String name, match, replace, match_var, replace_var;
   Regex ident("[a-zA-Z_][a-zA-Z0-9_]*");
 
-  if(verbose)
-    cerr<<"reading translate info in "<<i.filename()<<endl;
+    config.report(String("reading translate info in ")+i.filename(),
+		    configinfo::report_verbose);
   try{
-    do{
-      i.skip_space();
-      name=i.get_name(ident);
-      if(name=="")
-	i.skip_line();
-    } while (name=="");
-    if(debug)
-      cerr<<"name="<<name<<endl;
+    name=i.get_name(ident);
+    config.report(String("name=")+name,
+		      configinfo::report_debug);
     i.skip_space();
     match_var=i.get_name(ident);
-    if(debug)
-      cerr<<"match_var="<<match_var<<endl;
+    config.report(String("match_var=")+match_var,
+		    configinfo::report_debug);
     i.skip_space();
     i.skip_char('-');
     i.skip_char('>');
@@ -207,7 +310,7 @@ void translateinfo::init(parsestream &i){
       pair<const String, trans_class *> p;
 
       i.skip_space();
-      match=i.get_stringconst();
+      match=i.get_Stringconst();
       if(match==ENDTRANSLATE_TRANS){
 	i.skip_line();
 	break;
@@ -217,7 +320,7 @@ void translateinfo::init(parsestream &i){
 	continue;
       }
       i.skip_space();
-      replace=i.get_stringconst();
+      replace=i.get_Stringconst();
       (String)p.first=match;
       if(name=TRANSLATE_TRANS)
 	p.second=new translate(match,replace,replace_var);
@@ -230,28 +333,16 @@ void translateinfo::init(parsestream &i){
     }
   }
   catch (endoffile){}
-  catch (ident_expected){
-    cerr<<"Identifier expected in translates file"<<endl;
-    exit(1);
-  }
-  catch (char_expected c){
-    cerr<<"\""<<c.c<<"\" expected in line "<<i.linenumber()
-	<<" in file "<<i.filename()<<endl;
-    exit(1);
-  }
 }
 
 translateinfo::translateinfo(parsestream &i){
   init(i);
 }
 translateinfo::translateinfo(const String &filename){
-
-  if(debug)
-    cerr<<"attempting to open "<<filename<<".. ";
+  config.report(String("attempting to open ")+filename+".. ",
+		configinfo::report_debug);
   parsestream ps(filename);
 
-  if(!ps.good())
-    throw ferror_read(filename);
   init(ps);
 }
 
@@ -262,8 +353,8 @@ void translateinfo::process(menuentry &m){
   for(i=trans.begin(); i!=trans.end(); i++){
     match=&m.data[(*i).first];
     j=(*i).second.upper_bound(*match);
-    if(debug)
-      cerr<<"translate: var["<<*match<<"]"<<endl;
+    config.report(String("translate: var[")+*match+"]",
+			 configinfo::report_debug);
     do{
       j--;
       (*j).second->process(m,*match);
@@ -276,13 +367,13 @@ void translateinfo::debuginfo(){
   map<String, trans_map, less<String> >::iterator i;
   trans_map::iterator j;
   for(i=trans.begin(); i!=trans.end(); i++){
-    cerr<<"TRANS: ["<<(*i).first<<"]"<<endl;
+  config.report(String("TRANS: [")+(*i).first+"]",
+		configinfo::report_debug);
     for(j=(*i).second.begin();
 	j!=(*i).second.end();
 	j++){
-      cerr<<"key="<<(*j).first;
+      config.report(String("key=")+(*j).first,configinfo::report_debug);
       (*j).second->debuginfo(cerr);
-      cerr<<endl;
     }
   } 
 }
@@ -299,8 +390,8 @@ void read_pkginfo (const char *statusfile)
   
   if(!(status = fopen (statusfile, "r")))
     throw ferror_read(statusfile);
-  if (verbose)
-    cerr<<"Reading installed packages...\n"<<endl;
+  config.report(String("Reading installed packages..."),
+		configinfo::report_verbose);
   while (!feof (status)){
     fgets (tmp, MAX_LINE, status);
     if (strlen (tmp) == 0)
@@ -333,42 +424,42 @@ void read_menufile(const char *filename,
 		   const String &shortfilename,
 		   vector<String> &menudata,
 		   vector<String> &menudata_compat){
-  ifstream f(filename);
-  parsestream i(f);
+  bool wrote_filename=false;
+  parsestream *i;
+  procbuf pr;
+  istream *pipe_istr;
 
-  if(!f)
-    throw ferror_read(filename);
-  
-  if(debug){
-    cerr<<"Reading menuentryfile "<<filename<<endl;
-  }
+  config.report(String("Reading menuentryfile ")+filename,
+		configinfo::report_debug);
   try{
-    while(1){
+    if(executable(filename)){
+      pr.open(filename,ios::in);
+      pipe_istr=new istream(&pr);
+      i=new parsestream(*pipe_istr);
+    }
+    else
+      i=new parsestream(filename);
+    while(i->good()){
       try{
-	menuentry m(i,shortfilename);
+	menuentry m(*i,shortfilename);
 	if(transinfo)
 	  transinfo->process(m);
+	if(!wrote_filename){
+	  menudata.push_back(String("!F ") + filename + "\n");
+	  wrote_filename=true;
+	}
 	m.output(menudata);
 	m.output_compat(menudata_compat);
       }
       catch(cond_inst_false){}
-      catch (ident_expected){
-	cerr<<"Identifier expected in line "<<i.linenumber()
-	    <<" in file "<<filename<<endl;
-	throw informed_fatal();
-      }
-      catch (char_expected c){
-	cerr<<"\""<<c.c<<"\" expected in line "<<i.linenumber()
-	    <<" in file "<<filename<<endl;
-	throw informed_fatal();
-      }
-      catch (unknown_cond_package){
-	cerr<<"unknown install condition (currently, only \"package\" is supported)"<<endl;
-	throw informed_fatal();
-      }
     }
+    if(executable(filename)){
+      delete pipe_istr;
+      pr.close();
+    }
+    delete i;
   }
-  catch(endoffile){};
+  catch(endoffile p){}
 }
 
 void read_menufilesdir(const String &dirname,
@@ -379,12 +470,12 @@ void read_menufilesdir(const String &dirname,
   DIR *dir;
   struct dirent *entry;
   String name;
-  if(verbose)
-    cerr<<"Reading menuentryfiles  in "<<dirname<<endl;
+  config.report(String("Reading menuentryfiles  in ")+dirname,
+		configinfo::report_debug);
   dir=open_dir_check(dirname);
   while((entry=readdir(dir))){
     name=String(entry->d_name);
-    if((name!="README")&&(name[0]!='.'))
+    if((name!="README")&&(name!="core")&&(name[0]!='.'))
       if(menufiles_processed.find(name)==menufiles_processed.end()){
 	menufiles_processed.insert(name);
 	name=dirname+name;
@@ -415,9 +506,8 @@ void run_menumethod(char *methodname,
     md=&menudata_compat;
   else
     md=&menudata;
-  if(verbose){
-    cerr<<"Udate-menus: Running method:"<<methodname<<endl;
-  }
+  config.report(String("Udate-menus: Running method:")+methodname,
+		configinfo::report_verbose);  
   strcat(methodname," -f --stdin ");
   if((p=popen(methodname, "w"))){
     signal(SIGPIPE,SIG_IGN);
@@ -438,8 +528,8 @@ void run_menumethoddir (const String &dirname,
   struct dirent *entry;
   char *s, tmp[MAX_LINE];
 
-  if(verbose)
-    cerr<<"running menu-methods in "<<dirname<<endl;
+  config.report(String("running menu-methods in ")+dirname,
+		configinfo::report_verbose);
   dir=open_dir_check(dirname);
   while ((entry = readdir (dir)) != NULL){
     if (!strcmp (entry->d_name, "README"))
@@ -473,12 +563,13 @@ int kill_other_updatemenus(){
     {
       ifstream f(UPMEN_LOCKFILE);
       if(!f){
-	if(verbose)
-	  cerr<<"Update-menus: No other update-menus programme running. Good."<<endl;
+	config.report(String("Update-menus: No other update-menus programme running. Good."),
+		      configinfo::report_verbose);
 	return 0;
       }
       parsestream p(f);
-      t=p.get_stringconst();
+      p.set_filename(UPMEN_LOCKFILE);
+      t=p.get_Stringconst();
     }
     s="/proc/"+t+"/exe";
     stat(&(s[0]),&st_other);
@@ -487,13 +578,16 @@ int kill_other_updatemenus(){
     if((st_self.st_dev==st_other.st_dev)&&
        (st_self.st_ino==st_other.st_ino)){
       int other_pid=Stringtoi(t);
-      if(verbose)
-	cerr<<"Update-menus: another update-menus is running (pid="<<other_pid<<"), killing it."<<endl;
+      config.report(String("Update-menus: another update-menus is running (pid=")
+		    +itoString(other_pid)+"), killing it."
+		    ,configinfo::report_verbose);
       kill(other_pid,SIGKILL);
       return 1;
     }else
-      if(verbose)
-	cout<<"Update-menus: A process with same pid as in "UPMEN_LOCKFILE" is running, but appears to have different executable image. Ignoring it."<<endl;
+      config.report(String("Update-menus: A process with same pid as in "
+			   UPMEN_LOCKFILE
+			   " is running, but appears to have different executable image. Ignoring it.")
+		    ,configinfo::report_verbose);
   }
   return 0;
 }
@@ -504,7 +598,9 @@ void create_lock(){
     ofstream of(UPMEN_LOCKFILE);
     of<<getpid();
     if(!of.good()){
-      cerr<<"Update-menu: cannot write to lockfile "UPMEN_LOCKFILE<<". Aborting."<<endl;
+      config.report("Update-menu: cannot write to lockfile "
+		    UPMEN_LOCKFILE". Aborting.",
+		    configinfo::report_normal);
       exit(1);
     }
   }
@@ -513,7 +609,8 @@ void create_lock(){
 void remove_lock(){
   if(!getuid()){
     if(unlink(UPMEN_LOCKFILE))
-      cerr<<"Update-menus: Cannot remove lockfile "UPMEN_LOCKFILE<<endl;
+      config.report("Update-menus: Cannot remove lockfile "UPMEN_LOCKFILE,
+		    configinfo::report_normal);
   }
 }
 
@@ -575,8 +672,8 @@ void wait_dpkg(){
     }else{
       create_lock();
       if((!other)||verbose){
-        cerr<<"Update-menus: waiting for dpkg to finish (forking to background)"<<endl
-            <<"Update-menus: (checking "DPKG_LOCKFILE")"<<endl;
+	cerr<<"Update-menus: waiting for dpkg to finish (forking to background)"<<endl
+	    <<"Update-menus: (checking "DPKG_LOCKFILE")"<<endl;
       }
       while(check_dpkglock())
 	sleep(2);
@@ -593,9 +690,9 @@ void wait_dpkg(){
 void parse_params(char **argv){
   while(*(++argv)){
     if(String("-d")==*argv)
-      debug=verbose=1;
+      config.set_verbosity(configinfo::report_debug);
     if(String("-v")==*argv)
-      verbose=1;
+      config.set_verbosity(configinfo::report_verbose);
     if(String("-h")==*argv){
       cerr<<"update-menus: update the various window-manager config files (and"<<endl
 	  <<"  dwww, and pdmenu) after menuentry files were changed."<<endl
@@ -606,26 +703,53 @@ void parse_params(char **argv){
     }
   }
 }
+void read_userconfiginfo(){
+  //Because of gcc-2.7.2.1 internal compiler errors, had to split this up.
+  //(well, not sure of the configinfo stuff, but translateinfo failed)
+  if(getuid()){
+    try{
+      config.update(String(getenv("HOME"))+"/"+USERCONFIG);
+    } catch(ferror_open d){};
+  };
+}
+void read_rootconfiginfo(){
+  if(!transinfo){
+    try{
+      config.update(CONFIG_FILE);
+    }catch (ferror_open d){};
+  }
+}
+
+void read_usertranslateinfo(){
+  //Because of gcc-2.7.2.1 internal compiler errors, had to split this up.
+  if(getuid()){
+    try{
+      transinfo=new translateinfo(String(getenv("HOME"))+"/"+USERTRANSLATE);
+    } catch(ferror_open d){};
+  };
+}
+void read_roottranslateinfo(){
+  if(!transinfo){
+    try{
+      transinfo=new translateinfo(TRANSLATE_FILE);
+    }catch (ferror_open d){};
+  }
+}
 main (int, char **argv){
   vector<String> menudata, menudata_compat;
   String dummy;
   
   debug=0;
-  parse_params(argv);
-  wait_dpkg();
-  read_pkginfo(DPKG_STATUSFILE);
-  transinfo=NULL;
   try{
-    if(getuid()){
-      try{
-	transinfo=new translateinfo(String(getenv("HOME"))+"/"+USERTRANSLATE);
-      } catch(ferror_read d){dummy=d.name;}
-    };
-    if(!transinfo){
-      try{
-	transinfo=new translateinfo(TRANSLATE_FILE);
-      }catch (ferror_read d){dummy=d.name;};
-    }
+    //  read_userconfiginfo();
+    read_rootconfiginfo();
+    parse_params(argv);
+    wait_dpkg();
+    read_pkginfo(DPKG_STATUSFILE);
+    transinfo=NULL;
+
+    read_usertranslateinfo();
+    read_roottranslateinfo();
     if(debug)
       transinfo->debuginfo();
     if(getuid()){
@@ -654,15 +778,22 @@ main (int, char **argv){
     } else
       run_menumethoddir(MENUMETHODS, menudata, menudata_compat);
   }
-  catch (ferror_read f){
-    cerr<<"Cannot open file "<<f.name<<" for reading"<<endl;
-  }
-  catch (dir_error_read d){
-    cerr<<"Cannot open directory "<<d.name<<" for reading"<<endl;
-  }
-  catch (informed_fatal){
-    cerr<<"Aborting."<<endl;
-  }
+  catch(endoffile p){        p.report(cerr); }
+  catch(endofline p){        p.report(cerr); }
+  catch(char_expected p){    p.report(cerr); }
+  catch(char_unexpected p){  p.report(cerr); }
+  catch(def_error p){        p.report(cerr); }
+
+  catch(unknown_cond_package p){  p.report(cerr); }
+  //  catch (ferror_read f){
+  //    cerr<<"Cannot open file "<<f.name<<" for reading"<<endl;
+  //  }
+  //  catch (dir_error_read d){
+  //    cerr<<"Cannot open directory "<<d.name<<" for reading"<<endl;
+  //  }
+  //catch (informed_fatal){
+  //  cerr<<"Aborting."<<endl;
+  //  }
 
   remove_lock();
 }
